#ifndef HQS_IPP
#define HQS_IPP

namespace hqs {

/**
 * @brief Constructs an AIG from a single clause.
 *
 * @param clause the clause which is to be transformed into an And-Inverter graph.
 * @returns the constructed AIG
 */
template <typename Container>
aigpp::EdgeRef
HenkinSolver::clause2aig(const Container& clause)
{
    if (clause.empty()) return _manager.getConst0();

    _aig_buffer.clear();
    _aig_buffer.reserve(clause.size());

    std::transform(clause.cbegin(), clause.cend(), std::back_inserter(_aig_buffer),
                   [this](const Literal lit) -> aigpp::EdgeRef { return lit2aig(lit); });

    const auto result = aigpp::multiOr(_manager, _aig_buffer);
    _aig_buffer.clear();

    return result;
}

inline const Settings&
HenkinSolver::settings() const noexcept
{
    return _settings;
}

inline Settings&
HenkinSolver::settings() noexcept
{
    return _settings;
}

inline const Statistics&
HenkinSolver::statistics() const noexcept
{
    return _statistics;
}

inline aigpp::EdgeRef
HenkinSolver::var2aig(const Variable var) const
{
    const auto found = _var2aig.find(var);
    val_assert(found != _var2aig.cend() && found->second.isValid());
    return found->second;
}

inline Variable
HenkinSolver::aig2var(const aigpp::EdgeRef& aig) const
{
    return aig2var(aig.variableIndex());
}

inline Variable
HenkinSolver::aig2var(const unsigned int aig) const
{
    const auto found = _aig2var.find(aig);
    val_assert(found != _aig2var.cend());
    return found->second;
}

inline aigpp::EdgeRef
HenkinSolver::lit2aig(const Literal lit) const
{
    return var2aig(hqspre::lit2var(lit)).notIf(hqspre::isNegative(lit));
}

}  // end namespace hqs

#endif
